SystemOrganization addCategory: #'MQTT-Client'!

Object subclass: #MQTTClient
	instanceVariableNames: 'brokerID port sockStrm keepAliveTime readProcess lastMID pendingJobs willTopic willMessage willRetainFlag willQos cleanSessionFlag userNameString passwordString outgoingPacketQueue outgoingPacketProcess incomingPacketQueue incomingPacketProcess pingOK retryTime retryProcess currentSubscriptions clientID'
	classVariableNames: 'Debug'
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTClient commentStamp: 'tpr 3/16/2017 14:55' prior: 0!
An MQTTClient provides a connection to an MQTT broker. 

Google 'MQTT' & 'MQTT paho' for lots of information. Many comments through the code refer to sections wiinth the OASIS standards document 'docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html'

See the swiki page http://wii.squeak.org/squeak/63 for further documentation.

You will need to know the name or IP of a suitable MQTT broker (Google 'mosquitto' for a popular one) installation, and to have the broker running. You will want the mosquitto_client package too, so you can do command line publishing to test MQTTClient.

To play with the client try inspecting 
"MQTTClient new"
and then in the inspector paste and doit -
"
self connectTo: 'mymqttbrokerservername' port: nil.
self onTopic: 'mytestingtopic' do: [:t :m|
	Transcript show: 'got topic ', t, ' message ', m; cr].
"
This will show the topic and message content of any MQTT messages matching whatever you choose as 'mytestingtopic' on your Transcript. The wildcards '#' and '+' are handled and should behave as expected from the descriptions in the OASIS doc. To send messages you can use the mosquitto_pub command (from the above mentioned clients package) ie
	mosquitto_pub -h mymqttbrokerservername -t mytestingtopic -m "why hello mqtt"

When you want to unsubscribe from a topic use 
"
self unsubscribeFrom: 'mytestingtopic'.
"
And to disconnect cleanly use 
"
self disconnect
"

The Closure used as the second argument in the onTopic:do: message takes two parameters, the topic and the actual message. Your application would normally provide a Closure that performs the appropriate actions with the received message and topic. It would be possible to subscribe to a range of topics and then have your application work out what action is required from the specific topic, or to subscribe to many specific topics with a suitable Closure for each case. Note that the message is not in any way guaranteed to be a String; the data is transmitted as a byte array and could be an image, a string, some executable code to store away, pretty much anything.

Instance variables and their usages.

brokerID
	- the name or IP address (as a string) of the broker to connect to.
port
	- nil or a specific port number. Normally 1883 is used but 8883 can be used for TLS communication.
	
sockStrm
	- a SocketStream is used to provide the socket link to the broker

keepAliveTime
	- time in seconds for the keep alive process

readProcess
	- a process to sit on the socket stream and read packets as they come in.
	
lastMID
	- last used 16bit number for message ID use. We are supposed to track in-use IDs to avoid using the sameone twice simultaneously.

pendingJobs
	- Dictionary of pending job objects - a SUBACK following a SUBSCRIBE for example
	
willTopic/willMessage/willRetainFlag/willQos - the will message that gets sent to subscribers if the connection fails

cleanSessionFlag - if this is true the connection will represent a new session. Normally it would be set to false if trying reconnect after a problem.

userNameString/passwordString - UID & PWD for some brokers

outgoingPacketQueue/outgoingPacketProcess - SharedQueue and process to write packets we send out

incomingPacketProcess - process to read incoming packets and handle them

pingOK - flag for the keep-alive cycle; set false when we send a PINGREQ and back to true when we get a PINGRESP; tested to see if we're getting through

retryTIme/retryProcess - process that checks pending jobs for any that need to resend their packet; also fires the PINGREQ stuff at need

currentSubscriptions - list of MQTTSubscriptions that match topic subscriptions and action blocks


!

!MQTTClient class methodsFor: 'debugging' stamp: 'tpr 2/2/2017 17:06'!
debugOff
	"MQTTClient debugOff"
	Debug := false! !

!MQTTClient class methodsFor: 'debugging' stamp: 'tpr 2/2/2017 17:05'!
debugOn
	"MQTTClient debugOn"
	Debug := true! !

!MQTTClient class methodsFor: 'class initialization' stamp: 'tpr 2/2/2017 16:44'!
initialize
	"MQTTClient initialize"
	Debug := false! !

!MQTTClient methodsFor: 'pending jobs' stamp: 'tpr 2/3/2017 17:34'!
addPending: mqttJobToComplete
	"Add a record of a pending job that needs completion"
	pendingJobs ifNil:[pendingJobs := OrderedCollection new].
	self debugLog: 'add pending: ', mqttJobToComplete asString.
	pendingJobs add: mqttJobToComplete! !

!MQTTClient methodsFor: 'errors' stamp: 'tpr 3/16/2017 17:30'!
badMessageError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTClient methodsFor: 'errors' stamp: 'tpr 2/24/2017 16:58'!
badPacketID

	self halt: 'Unable to find acceptable packet ID for new packet'! !

!MQTTClient methodsFor: 'errors' stamp: 'tpr 2/24/2017 14:33'!
badTopicError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTClient methodsFor: 'errors' stamp: 'tpr 2/24/2017 17:18'!
badWill

	self halt: 'Bad Will topic, message etc'! !

!MQTTClient methodsFor: 'accessing' stamp: 'tpr 3/1/2017 16:49'!
clientID: clientIDString
	"set the id that the server will use to identify the session being established by this connect packet. It ought to be unique within the scope of this image, ideally totally unique. That's up to the user of this class though"
	clientID := clientIDString! !

!MQTTClient methodsFor: 'accessing' stamp: 'tpr 3/1/2017 16:48'!
clientIDString
	"ought to come from the client but if it hasn't been set use a vaguely useful default"
	^clientID ifNil:['SqueakMQTT']! !

!MQTTClient methodsFor: 'connecting' stamp: 'tpr 8/2/2017 12:59'!
connect
	"connect the socket stream to the broker's port, send my CONNECT packet and fire up the supporting processes for reading, writing and retry"
	
	self initializeSocketStream.

	self initializeMQTTConnection.
	
	self initializeKeepAlive;
		initializePacketReading;
		initializePacketWriting;
		initializeRetryProcess! !

!MQTTClient methodsFor: 'public api' stamp: 'tpr 8/2/2017 13:00'!
connectTo: aHostName port: aPortOrNil
	"set the brokerID and port, then connect. Any already set retain, will or keep-alive settings will be used"

	self host: aHostName port: aPortOrNil.
	self connect! !

!MQTTClient methodsFor: 'debugging' stamp: 'tpr 2/2/2017 16:43'!
debugLog: aString
	Debug ifTrue: [
		Transcript show: aString; cr]
	! !

!MQTTClient methodsFor: 'accessing' stamp: 'tpr 1/26/2017 14:42'!
defaultPort
	"the default for MQTT is 1883. 8883 is used for SSL connections"
	^1883! !

!MQTTClient methodsFor: 'public api' stamp: 'tpr 7/31/2017 16:32'!
disconnect
	"close the connection"
	self debugLog: 'disconnecting'.
	incomingPacketProcess ifNotNil:[:pp| pp terminate].
	outgoingPacketProcess ifNotNil:[:pp| pp terminate].
	retryProcess ifNotNil:[:pp| pp terminate].
	sockStrm ifNotNil: [
		[MQTTPacketDisconnect new encodeOn: sockStrm] on: NetworkError do:[:ex| Debug ifTrue:[ex pass]]. "ignore unless we're in debug mode"
		sockStrm close].
	sockStrm := incomingPacketProcess := outgoingPacketProcess := retryProcess := nil! !

!MQTTClient methodsFor: 'subscribing' stamp: 'tpr 2/7/2017 15:06'!
doSubscriptionActionsFor: aPublishPacket
	"find the subscriptions matching the topic of this packet and the corresponding Blocks to evaluate, then run them"
	currentSubscriptions do: [ :s|
		s handlePacket: aPublishPacket]! !

!MQTTClient methodsFor: 'pending jobs' stamp: 'tpr 3/16/2017 17:38'!
findAndRunJobsOverdueForAcknowledgement
	"Find and process the pending job(s) that needs to resend a packet"
	|  tick |
	pendingJobs ifNil:[^nil].
	tick := Time totalSeconds - retryTime. 
	"looks odd but works as a way to handle possible changes in retryTime - 
	we actually check if the last send time is more than retryTime seconds ago"
	pendingJobs do: [:j| j resendFor: self ifNeededAtTime: tick]! !

!MQTTClient methodsFor: 'pending jobs' stamp: 'tpr 3/17/2017 17:37'!
findPendingJob: testBlock
	"Find the matching record of a pending job that needs completion. Your testBlock should be written to find only a single result but we use #detect: here and only give you back the first one"

	pendingJobs ifNil:[^nil].
	^pendingJobs detect: [:e| testBlock value: e] ifNone: [nil]! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 1/27/2017 14:35'!
handleIncomingPacket: anMQTTPacket
	"most of the packets that come to me are acknowledgements of one sort or another; the exception is the PUBLISH, which we have to handle since it is how we get that pesky data.
	Now most people might be expecting some sort of case statement here, switching based on the type of the packet. But we're Smalltalkers and that is more than a little gauche when we can just do it right and delegate to the packet"
	
	anMQTTPacket evaluateFor: self! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 2/3/2017 11:42'!
handlePingRespPacket: aPingRespPacket
	"just set the pingOK status to true"

	self debugLog: 'rx PINGRESP for: ', aPingRespPacket .
	pingOK := true! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 2/6/2017 17:01'!
handlePubAckPacket: aPubAckPacket
	"The broker has confirmed a qos=1 publish process (3.4)"
	| job |

	"Find my pending job by matching the msgID tag"
	self debugLog: 'rx PUBACK for: ', aPubAckPacket packetID.
	job := self findPendingJob:[:j|
		j isPubAckID: aPubAckPacket packetID].
	job ifNil: [^ self missingPendingJob].
	"What do we do?"
	job completeFor: self.
	"c) consider the publish completed. Not sure what that means in practice"! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 2/24/2017 16:42'!
handlePubCompPacket: aPubCompPacket
	"The broker has confirmed a qos=2 publish process is completed (3.7)"
	| job |

	"Find my pending job by matching the msgID tag"
	self debugLog: 'rx PUBCOMP for: ', aPubCompPacket packetID.
	job := self findPendingJob: [:j|
		j isPubCompID:  aPubCompPacket packetID].
	job ifNil: [^ self missingPendingJob].
	"What do we do?"
	job completeFor: self.
	"c) consider the publish completed. Not sure what that means in practice"! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 2/24/2017 16:42'!
handlePubRecPacket: aPubRecPacket
	"The broker has confirmed a qos=2 publish process is in progress (3.5)"
	| job pubRelPacket |

	"Find my pending job by matching the msgID tag"
	self debugLog: 'rx PUBREC for: ', aPubRecPacket packetID.
	job := self findPendingJob: [:j|
		j isPubRecID: aPubRecPacket packetID].
	job ifNil: [^ self missingPendingJob].
	"We need to respond by sending out a PUBREL packet with the *same* packetID, and add a pending job to handle the closing PUBCOMP"
	self sendPacket: (pubRelPacket := aPubRecPacket acknowledgement).
	"release the pending job from the list but do *not* release the packetID yet"
	job completeFor: self.
	"add a new pending job for the response to the PUBREL"
	self addPending: pubRelPacket pendingJob! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 2/24/2017 16:42'!
handlePubRelPacket: aPubRelPacket
	"The broker has confirmed a qos=2 publish process is completing, so accept the end of this relationship gracefully (3.6)"
	| job  |

	"Find my pending job by matching the msgID tag"
	self debugLog: 'rx PUBREL for: ', aPubRelPacket packetID.
	job := self findPendingJob: [:j|
		j isPubRelID: aPubRelPacket packetID].
	job ifNil: [^ self missingPendingJob].
	"We need to respond by sending out a PUBCOMP packet with the *same* packetID"
	self sendPacket:  aPubRelPacket acknowledgement.
	"release the pending job from the list and release the packetID"
	job completeFor: self
! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 2/24/2017 16:43'!
handlePublishPacket: aPublishPacket qos: packetQos
	"The broker has sent me a PUBLISH packet (3.3); extract the data and..."

	"Find my pending job by matching the msgID tag"
	self debugLog: 'rx PUBLISH with: ', aPublishPacket asString.
	"What do we do?"
	"qos = 0 ifTrue: [do nothing]."
	packetQos >= 1 ifTrue: ["send back an appropriate PUBACK or PUBREC packet"
		self sendPacket: aPublishPacket acknowledgement].
	packetQos = 2 ifTrue: ["more complex. 
		see if there is already a pending job to deal with a PUBREL
			if so, we're done, return
			if not, add a pending PUBREL job and pass on the data"
		(self findPendingJob: [:j|
		j isPubRelID: aPublishPacket packetID]) 
			ifNotNil: [^self].
		self addPending: aPublishPacket pendingReleaseJob].

	"and here we do whatever to let the user have the data"
	self doSubscriptionActionsFor: aPublishPacket! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 2/24/2017 16:44'!
handleSubAckPacket: aSubAckPacket
	"The broker has confirmed a SUBSCRIBE from me (3.9)"
	| job |

	"Find my pending job by matching the msgID tag"
	self debugLog: 'rx SUBACK for: ', aSubAckPacket packetID.
	job := self findPendingJob: [:j|
		j isSubAckID: aSubAckPacket packetID].
	job ifNil: [^ self missingPendingJob].
	job completeFor: self
! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 2/24/2017 16:44'!
handleUnsubAckPacket: anUnsubAckPacket
	"The broker has confirmed an UNSUB process (3.10)"
	| job |

	"Find my pending job by matching the msgID tag"
	self debugLog: 'rx UNSUBACK for: ', anUnsubAckPacket packetID.
	job := self findPendingJob: [:j|
		j isUnsubAckID: anUnsubAckPacket packetID].
	job ifNil: [^ self missingPendingJob].
	job completeFor: self! !

!MQTTClient methodsFor: 'initialize-release' stamp: 'tpr 3/16/2017 17:09'!
host: aHostName port: aPortOrNil
	"set the brokerID and port. 
	The host name will be validated when we try to connect, so no checking is needed here.
	The port can be nil, in which case it will default to 1883"

	brokerID := aHostName.
	aPortOrNil ifNil:[port := self defaultPort] ifNotNil: [port := aPortOrNil]! !

!MQTTClient methodsFor: 'initialize-release' stamp: 'tpr 2/24/2017 16:46'!
initialize
	keepAliveTime := 0. "a sensible default that tells the broker to just assume we're always connected. Imagine the ways that can go wrong with poor connections... "
	retryTime := 2.
	cleanSessionFlag := true. 
	currentSubscriptions := OrderedCollection new:4! !

!MQTTClient methodsFor: 'keep alive' stamp: 'tpr 2/6/2017 16:20'!
initializeKeepAlive
	"set up the keep alive process. If the keepalive time is 0 the server should simply leave the connection open as long as possible and so we don't worry about this process. Otherwise, add a pending ping job that will repeat"
	keepAliveTime = 0 ifFalse: [self addPending: MQTTPendingPingJob new]! !

!MQTTClient methodsFor: 'connecting' stamp: 'tpr 3/15/2017 14:23'!
initializeMQTTConnection
	"assemble a CONNECT packet and send it"
	|connectPacket connAckPacket|
	
	connectPacket := MQTTPacketConnect new keepAliveTime: keepAliveTime.

	"do I want to set the will data?"						
	(willTopic isString and: [willMessage isString] and: [willRetainFlag isBoolean] and: [willQos isInteger]) 
		ifTrue: [connectPacket willTopic: willTopic message: willMessage retain: willRetainFlag qos: willQos].

	"set the clean session flag; the default for a new client connection is true but can be set differently; typically
	after a socket failure and in preparation for a re-connect attempt"
	connectPacket cleanSession: cleanSessionFlag.

	"do I need a username&|password? Nils are handled by the packet"
	connectPacket user: userNameString password: passwordString.
	
	"set the clientID by which the server knows us - see 3.1.3
	It defaults to SqueakMQTT but should be set separately by each user to keep it unique"
	connectPacket clientID: self clientIDString.

	self debugLog: 'sending ', connectPacket asString.
	connectPacket encodeOn: sockStrm. "<---------------------- seems like a place to wrap in an exception handler, but what to do?"
	connAckPacket := MQTTPacket readFrom: sockStrm.
	self assert: [connAckPacket class = MQTTPacketConnAck].
	self assert: [connAckPacket sessionPresent isKindOf:Boolean]
! !

!MQTTClient methodsFor: 'initialize-release' stamp: 'tpr 2/24/2017 15:59'!
initializePacketReading
	"set up the shared queue and process to write packets. The
	connection must be open"
	incomingPacketProcess := [| packet |
			[packet := [MQTTPacket readFrom: sockStrm] 
						on: NetworkError
						do: [:e | 
							self debugLog: e.
							self reconnect.
							e retry].
			self handleIncomingPacket: packet] repeat]
				forkNamed: 'MQTT packet reading'! !

!MQTTClient methodsFor: 'initialize-release' stamp: 'tpr 2/13/2017 11:59'!
initializePacketWriting
	"set up the shared queue and process to write packets. The connection must be open"
	outgoingPacketQueue := SharedQueue new: 8.
	outgoingPacketProcess := 
		[[ |packet| 
				packet := outgoingPacketQueue next.
				[packet  encodeOn: sockStrm]
					on: NetworkError do: [:e|
						self debugLog: e.
						self reconnect.
						e retry]] repeat ] forkNamed: 'MQTT packet writing'! !

!MQTTClient methodsFor: 'keep alive' stamp: 'tpr 3/16/2017 17:38'!
initializeRetryProcess
	"some packets may need resending if they don't get acknowledged"
	retryProcess := [|loopDelay|
				loopDelay := Delay forSeconds: 1.
				[loopDelay wait.
				self findAndRunJobsOverdueForAcknowledgement] repeat
		] forkNamed: 'MQTTClient retry loop'! !

!MQTTClient methodsFor: 'connecting' stamp: 'tpr 3/15/2017 17:17'!
initializeSocketStream
	"make the basic socket connection to the broker"

	"Possible errors here:-
	SocketPrimitiveFailed - too serious to try to fix here.
	NoNetworkError - failed to be able to find any network interface. Really serious, can't handle here.
	NameLookupFailure - brokerID is bad or it's not reachable. Could allow changing of broker ID or simply hold until something is fixed.
	ConnectionTimedOut or ConnectionRefused - some problem with making a connection. Could allow retrying."
	
	
	sockStrm ifNotNil: [sockStrm destroy].		
	[sockStrm := SocketStream openConnectionToHostNamed: brokerID port: port]
			on: NameLookupFailure 
			do: [:ex| (MQTTBrokerLookupFailure brokerName: ex hostName) signal.
					ex retry]
			on: ConnectionTimedOut, ConnectionRefused
			do: [:ex| (MQTTConnectionException signal: 'socket connection failed').
				ex retry].
 
	sockStrm binary; noTimeout! !

!MQTTClient methodsFor: 'public api' stamp: 'tpr 2/6/2017 16:32'!
keepAliveTime: timeInSecs
	"set the keep-alive time, in seconds not mS, that should be used by the server and the keepalive process. Do not alter this after the connection is opened"
	keepAliveTime := timeInSecs
! !

!MQTTClient methodsFor: 'errors' stamp: 'tpr 1/27/2017 15:09'!
missingPendingJob
	"I couldn't find a matching pending job. oops."
	self halt: 'No matching pending job'! !

!MQTTClient methodsFor: 'accessing' stamp: 'tpr 2/24/2017 22:14'!
newPacketID
	"provide a suitable 16 bit numeric packet id.
	This is supposed to be a unique number, presumably across the field of curently pending jobs.
	This ought to find a sutable value quickly in most cases, since we don't expect to have many pending jobs. If we get back to the 
	value that was last used then we must have gone all the way round and must fail bcause we're out of IDs"

	|prevMID|
	lastMID ifNil:[lastMID := 0].
	prevMID := lastMID.
	[lastMID := lastMID + 1 \\ 16rFFFF.
	pendingJobs ifNil: [^lastMID] ifNotNil: [:pj| pj detect: [:jb| jb packetID = lastMID] ifNone: [^lastMID]].
	prevMID = lastMID] whileFalse.

	^self badPacketID! !

!MQTTClient methodsFor: 'public api' stamp: 'tpr 2/7/2017 17:49'!
onTopic: topicString do: aBlock
	"a basic subscribe and do something message."
	self onTopic: topicString qos: 0 do: aBlock! !

!MQTTClient methodsFor: 'public api' stamp: 'tpr 2/24/2017 17:03'!
onTopic: topicString qos: qos do: aBlock
	"a basic subscribe and do something message. We must check the topicString's acceptability and fail if there are issues"
	|subscription|
	subscription := MQTTSubscription for: topicString do: aBlock.
	subscription ifNil:[^self badTopicError].
	currentSubscriptions add: subscription.
	self subscribe: topicString qos: qos! !

!MQTTClient methodsFor: 'keep alive' stamp: 'tpr 2/6/2017 16:16'!
ping
	"send a PINGREQ and scehdule a pending PINGRESP. Return a time for the next ping"
	pingOK := false.
	self sendPacket:  MQTTPacketPingReq new.
	^Time totalSeconds + keepAliveTime
! !

!MQTTClient methodsFor: 'public api' stamp: 'tpr 2/3/2017 16:49'!
publishTopic: aTopic message: msgString qos: qos
	"publish the msgString to the connected broker. If qos > 0 we'll need to schedule a pending job for the ack sequence(s)"
	|pubPacket|
	pubPacket := MQTTPacketPublish new
						topic: aTopic message: msgString asByteArray;
						messageID: self newPacketID;
						qos: qos.
	qos > 0 ifTrue:[
		self addPending: pubPacket pendingJob].
	self sendPacket:  pubPacket! !

!MQTTClient methodsFor: 'connecting' stamp: 'tpr 2/24/2017 17:09'!
reconnect
	"reconnect the socket stream to the broker's port, send my CONNECT packet"
	
	self initializeSocketStream.

	cleanSessionFlag := false. "reconnecting, so we would prefer the broker to return us to prior condition where possible"
	self initializeMQTTConnection! !

!MQTTClient methodsFor: 'pending jobs' stamp: 'tpr 2/3/2017 18:18'!
releasePendingJob: pendingJob
	"Find the matching record of a pending job and remove it from the list"
	pendingJobs ifNil:[^nil].
	^pendingJobs remove: pendingJob  ifAbsent:[ nil]! !

!MQTTClient methodsFor: 'handling packets' stamp: 'tpr 2/24/2017 17:11'!
sendPacket: anMQTTPacket
	"add the packet to the outgoing queue where it will get sucked into The Machine as soon as possible"
	self debugLog: 'sending ', anMQTTPacket asString.
	outgoingPacketQueue nextPut: anMQTTPacket! !

!MQTTClient methodsFor: 'subscribing' stamp: 'tpr 2/24/2017 17:13'!
subscribe: aTopic qos: qos
	"set up a subscription to aTopic with the broker. Use #onTopic:do: for actual client applications"
	|subPacket|
	subPacket := MQTTPacketSubscribe new
						addTopic: aTopic qos: qos;
						packetID: self newPacketID.
	self addPending: subPacket pendingJob.
	self sendPacket:  subPacket
! !

!MQTTClient methodsFor: 'subscribing' stamp: 'tpr 2/7/2017 17:24'!
unsubscribe: aTopic
	"unsubscribe from aTopic and scehdule a pending job to handle the UnsubAck"
		|unsubPacket|
	unsubPacket := MQTTPacketUnsubscribe new
						addTopic: aTopic;
						packetID: self newPacketID.
	self addPending: unsubPacket pendingJob.
	self sendPacket:  unsubPacket
! !

!MQTTClient methodsFor: 'public api' stamp: 'tpr 2/7/2017 17:55'!
unsubscribeFrom: aTopic
	"unsubscribe from aTopic - remove the subscription from currentSubscriptions and then tell the broker to unsubscribe"

	currentSubscriptions removeAllSuchThat: [:s| s topicMatches: aTopic].
	self unsubscribe: aTopic! !

!MQTTClient methodsFor: 'public api' stamp: 'tpr 1/26/2017 13:00'!
username: uName password: pwd
	"set the MQTT user name and password to be used by the connection. Nils are ok for either or both"
	uName ifNotNil:[
		userNameString := uName asString].
	pwd ifNotNil: [
		passwordString := pwd asString]
	! !

!MQTTClient methodsFor: 'public api' stamp: 'tpr 3/1/2017 16:39'!
willTopic: topicString message: messageString retain: retainBoolean qos: qosValue
	"setup a will message. We must have a topic, the actual message, a retain flag and requested QOS"
	(topicString isString and: [messageString isString] and: [retainBoolean = true or: [retainBoolean = false]] and: [qosValue isInteger] and: [qosValue between: 0 and: 2]) ifFalse: [
		^self badWill].
	willTopic := topicString.
	willMessage := messageString.
	willRetainFlag := retainBoolean.
	willQos := qosValue! !

Object subclass: #MQTTPacket
	instanceVariableNames: ''
	classVariableNames: 'MQTTProtocolLevel311 PacketTypes'
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacket commentStamp: 'tpr 2/24/2017 17:20' prior: 0!
MQTTPacket is the root class for specific varieties of MQTT packet. It only provides the outline of the fixed header etc. Subclasses provide the packets that have only the fixed header - 
	DISCONNECT
	PINGREQ
	PINGRESP
	
Others provide the sub-hierarchies for packets with variable headers and payloads.

See the comment for MQTTClient for usage details!

!MQTTPacket class methodsFor: 'debugging' stamp: 'tpr 3/14/2017 11:59'!
badPacketTypeError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTPacket class methodsFor: 'debugging' stamp: 'tpr 3/14/2017 12:00'!
emptyStreamError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTPacket class methodsFor: 'instance creation' stamp: 'tpr 12/22/2016 17:00'!
initialize
	"set up the packet type list etc"
	"self initialize"
	PacketTypes := {MQTTPacketConnect. MQTTPacketConnAck. MQTTPacketPublish. MQTTPacketPubAck. MQTTPacketPubRec. MQTTPacketPubRel. MQTTPacketPubComp. MQTTPacketSubscribe. MQTTPacketSubAck. MQTTPacketUnsubscribe. MQTTPacketUnsubAck. MQTTPacketPingReq. MQTTPacketPingResp. MQTTPacketDisconnect.} .
	MQTTProtocolLevel311 := 4 "encoding for CONNECT etc"! !

!MQTTPacket class methodsFor: 'instance creation' stamp: 'tpr 12/15/2016 16:44'!
readFrom: aReadStream
"read the stream to work out which kind of packet it contains, create a suitable instance and get that to read the rest of the data"
	^aReadStream peek
		ifNil:[" we're out of luck, error" self emptyStreamError]
		ifNotNil:[ :byte|
			(PacketTypes at: byte >>4 ifAbsent:[self badPacketTypeError]) new readFrom: aReadStream]! !

!MQTTPacket methodsFor: 'errors' stamp: 'tpr 12/19/2016 21:16'!
badConnectAcknowledgeError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTPacket methodsFor: 'errors' stamp: 'tpr 12/15/2016 16:55'!
badFixedHeaderError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTPacket methodsFor: 'errors' stamp: 'tpr 1/24/2017 12:56'!
badQosError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTPacket methodsFor: 'errors' stamp: 'tpr 12/26/2016 17:44'!
badTopicError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTPacket methodsFor: 'errors' stamp: 'tpr 12/28/2016 14:05'!
badTopicListError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTPacket methodsFor: 'errors' stamp: 'tpr 12/19/2016 21:20'!
connectionRefusedError: refusalCode
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTPacket methodsFor: 'encoding' stamp: 'tpr 12/22/2016 16:43'!
decode16BitIntegerFrom: aReadStream
	"read a 16bit number MSB format from the stream"
	^aReadStream next  << 8 + aReadStream next! !

!MQTTPacket methodsFor: 'fixed header' stamp: 'tpr 2/3/2017 12:13'!
decodeFixedHeaderFrom: aReadStream
"read the fixed header from the stream and check it for decent state; the only class needing to do anything clever is MQTTPacketPublish"
	aReadStream next = self fixedHeader ifFalse:[ self badFixedHeaderError].
	self decodeLengthFrom: aReadStream! !

!MQTTPacket methodsFor: 'encoding' stamp: 'tpr 12/19/2016 21:09'!
decodeFrom: aReadStream
	"decode me from the stream. I only have a fixed header"
	self decodeFixedHeaderFrom: aReadStream! !

!MQTTPacket methodsFor: 'encoding' stamp: 'tpr 2/3/2017 12:11'!
decodeLengthFrom: aReadStream
"the default for fixed header only packets is 0 but we have to pull the next byte from the stream to keep the input buffer balanced"
	| byte |
	byte := aReadStream next.
	byte > 0 ifTrue: [self encodedLengthError]
	! !

!MQTTPacket methodsFor: 'encoding' stamp: 'tpr 12/22/2016 16:38'!
encode16BitInteger: smallNumber on: aWriteStream
	"write a 16bit number MSB format on the stream"
	aWriteStream nextPut: (smallNumber >>8); nextPut: (smallNumber bitAnd: 16rFF)! !

!MQTTPacket methodsFor: 'encoding' stamp: 'tpr 12/26/2016 12:19'!
encodeOn: aWriteStream
	"basic encoding of my data onto a stream (usually a socket) to transmit. I only have a fixed header.
	The fixed header only classes are all 2 bytes in size, giving a remaining length of 0.
	See MQTTPacketVariableHeaded>>#encodeOn: for the more complex arrangements for variable size packets"
	aWriteStream nextPut: self fixedHeader; nextPut: 0; flush.
	^aWriteStream! !

!MQTTPacket methodsFor: 'encoding' stamp: 'tpr 7/31/2017 16:06'!
encodeString: aString on: aWriteStream
	"generic write of string to the strea,; convert to UTF8, prepend with the 16 bit MSB length of said utf8 string"
	|bytes|
	bytes := aString squeakToUtf8 asByteArray.
	self encode16BitInteger: bytes size on: aWriteStream.
	aWriteStream nextPutAll: bytes! !

!MQTTPacket methodsFor: 'errors' stamp: 'tpr 12/15/2016 16:21'!
encodedLengthError
"raise an exception at some point; for now just halt"
	^self halt! !

!MQTTPacket methodsFor: 'handling' stamp: 'tpr 1/27/2017 14:34'!
evaluateFor: anMQTTCLient
	"I've been received by the client so now is the time to come to the aid of the party"
	^self subclassResponsibility! !

!MQTTPacket methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:31'!
fixedHeader
"return the byte for the fixed header; 
this is the packet type 4 bit value << 4 | the flags particular to the type.
For almost all kinds of packet this is a fixed number. "
	^self packetType << 4 bitOr: self fixedHeaderFlags! !

!MQTTPacket methodsFor: 'fixed header' stamp: 'tpr 1/26/2017 17:32'!
fixedHeaderFlags
	"default for most classes is 0"
	^0! !

!MQTTPacket methodsFor: 'encoding' stamp: 'tpr 12/24/2016 17:32'!
readFrom: aReadStream
	self decodeFrom: aReadStream! !

MQTTPacket subclass: #MQTTPacketDisconnect
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketDisconnect commentStamp: 'tpr 12/16/2016 18:12' prior: 0!
A MQTTPacketDisconnect is how we close down the server connection and is the final action before closing sockets etc. See section 3.14

There is no variable header nor payload. We must not send any more packets to the server after sending the DISCONNECT and must close the network connection to the server. This does NOT cause any associated Will Message to be published. (3.14.4)!

!MQTTPacketDisconnect methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:36'!
packetType
	^14! !

MQTTPacket subclass: #MQTTPacketPingReq
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketPingReq commentStamp: 'tpr 12/16/2016 18:08' prior: 0!
A MQTTPacketPingReq is a way for us to ping the server to make sure that the network is still awake, that the server is still talking to us and that we are still here. See section 3.12

There is no variable header nor payload. We expect a PINGRESP in return (3.12.4)
!

!MQTTPacketPingReq methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:36'!
packetType
	^12! !

MQTTPacket subclass: #MQTTPacketPingResp
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketPingResp commentStamp: 'tpr 12/16/2016 18:09' prior: 0!
A MQTTPacketPingResp gets sent back from the server in response to a PINGREQ. See section 3.13

There is no variable header nor payload.!

!MQTTPacketPingResp methodsFor: 'handling' stamp: 'tpr 1/27/2017 15:44'!
evaluateFor: anMQTTClient
	"The broker has responded to a PINGREQ from me; we should check that it is reasonably timely and if not, close the connection and do whatever retry stuff seems proper"
	anMQTTClient handlePingRespPacket: self! !

!MQTTPacketPingResp methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:36'!
packetType
	^13! !

MQTTPacket subclass: #MQTTPacketVariableHeaded
	instanceVariableNames: 'remainingLength msgID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketVariableHeaded commentStamp: 'tpr 12/16/2016 18:28' prior: 0!
MQTTPacketVariableHeaded is an abstract class for the packets that include a variable header.
	CONNACK
	PUBACK
	PUBCOMP
	PUBREC
	PUBREL
	UNSUBACK
!

MQTTPacketVariableHeaded subclass: #MQTTPacketConnAck
	instanceVariableNames: 'sessionPresent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketConnAck commentStamp: 'tpr 1/18/2017 16:23' prior: 0!
An MQTTPacketConnAck is returned by the server after we send a CONNECT, per MQTT3.1.1 doc.
We only receive these and a CONNACK is supposed to be the first packet sent from the serer after a CONNECT. (3.2).

The variable header has no msgID field but instead 2 bytes containing
a) connect acknowledge flags - (3.2.2.1) - currently just the lowest bit of the byte representing the Session Present flag
b) connect return code (3.2.2.3)- an accept (\0) or failure code (\1-\5 see table 3.1)

There is no payload
!

!MQTTPacketConnAck methodsFor: 'variable header' stamp: 'tpr 12/26/2016 13:01'!
decodeVariableHeaderFrom: aReadStream
	"I don't use the msgID stuff; my variable header contains a connect flag and the acknowledge flags per section 3.2.2 of the spec"
	| byte1 byte2 |
	byte1 := aReadStream next.
	(byte1 anyMask: 16rFE) ifTrue:[ ^self badConnectAcknowledgeError].
	sessionPresent := byte1 allMask: 1.	

	byte2 := aReadStream next.
	byte2 > 0 ifTrue: [ ^ self connectionRefusedError: byte2]

	"somebody will need to be told about the sessionPresent flag value"! !

!MQTTPacketConnAck methodsFor: 'variable header' stamp: 'tpr 12/25/2016 12:33'!
encodeVariableHeaderOn: aWriteStream
"write the connect acknowledge flags and return code to the stream; this is purely for tesint since we don't send these packets"
	aWriteStream nextPut: sessionPresent asBit; nextPut: 0! !

!MQTTPacketConnAck methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:34'!
packetType
	^2! !

!MQTTPacketConnAck methodsFor: 'variable header' stamp: 'tpr 1/13/2017 17:52'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' session present: '.
	sessionPresent printOn: aStream! !

!MQTTPacketConnAck methodsFor: 'accessing' stamp: 'tpr 1/18/2017 16:57'!
sessionPresent
	"after connecting the client should check that this is ok; a false value implies the server doesn't have any retained state from a prior connection. This may be important to the client"
	^sessionPresent! !

MQTTPacketVariableHeaded subclass: #MQTTPacketPubAck
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketPubAck commentStamp: 'tpr 12/16/2016 15:12' prior: 0!
An MQTTPacketPubAck is the response to a PUBLISH packet with a qos = 1. See section 3.4 We have to both send and receive them.

The variable header contains the msgID that was included in the relevant PUBLISH packet. There is no payload.!

!MQTTPacketPubAck methodsFor: 'fixed header' stamp: 'tpr 1/27/2017 15:47'!
evaluateFor: anMQTTClient
	"The broker has responded to a PUBLISH from me with qos=1"
	anMQTTClient handlePubAckPacket: self! !

!MQTTPacketPubAck methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:36'!
packetType
	^4! !

MQTTPacketVariableHeaded subclass: #MQTTPacketPubComp
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketPubComp commentStamp: 'tpr 12/16/2016 16:30' prior: 0!
A MQTTPacketPubComp is the final response to a PUBLISH with a qos = 2. We need to both send and receive these. See section 3.6

The variable header contains the msgID that was included in the relevant PUBLISH packet. There is no payload.
!

!MQTTPacketPubComp methodsFor: 'handling' stamp: 'tpr 1/27/2017 16:06'!
evaluateFor: anMQTTClient
	"The broker has responded to a PUBLISH from me with qos=2"
	anMQTTClient handlePubCompPacket: self! !

!MQTTPacketPubComp methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:37'!
packetType
	^7! !

MQTTPacketVariableHeaded subclass: #MQTTPacketPubRec
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketPubRec commentStamp: 'tpr 12/16/2016 15:13' prior: 0!
A MQTTPacketPubRec is the response to a PUBLISH packet with qos =2; it is the second part of the exchange. We have to bothe send and receive them. See section 3.5

The variable header contains the msgID from the relevant PUBLISH packet. There is no payload.!

!MQTTPacketPubRec methodsFor: 'pending jobs' stamp: 'tpr 2/4/2017 17:06'!
acknowledgement
	"return a PUBREL packet to acknowledge receiveing me"
	^MQTTPacketPubRel new packetID: msgID! !

!MQTTPacketPubRec methodsFor: 'handling' stamp: 'tpr 1/27/2017 16:05'!
evaluateFor: anMQTTClient
	"The broker has responded to a PUBLISH from me with qos=2"
	anMQTTClient handlePubRecPacket: self! !

!MQTTPacketPubRec methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:37'!
packetType
	^5! !

MQTTPacketVariableHeaded subclass: #MQTTPacketPubRel
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketPubRel commentStamp: 'tpr 12/16/2016 16:29' prior: 0!
A MQTTPacketPubRel is the third response to a PUBLISH with a qos = 2. We need to both send and receive these. See section 3.6

The variable header contains the msgID that was included in the relevant PUBLISH packet. There is no payload.!

!MQTTPacketPubRel methodsFor: 'handling' stamp: 'tpr 2/3/2017 13:02'!
acknowledgement

	^MQTTPacketPubComp new packetID: msgID ! !

!MQTTPacketPubRel methodsFor: 'handling' stamp: 'tpr 1/27/2017 16:07'!
evaluateFor: anMQTTClient
	"The broker has responded to a PUBLISH from me with qos=2"
	anMQTTClient handlePubRelPacket: self! !

!MQTTPacketPubRel methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:45'!
fixedHeaderFlags
	^2! !

!MQTTPacketPubRel methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:37'!
packetType
	^6! !

!MQTTPacketPubRel methodsFor: 'pending jobs' stamp: 'tpr 2/4/2017 17:08'!
pendingJob
	"Return a pending PUBCOMP to complete later"
	^MQTTPendingPubCompJob for: self! !

MQTTPacketVariableHeaded subclass: #MQTTPacketUnsubAck
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketUnsubAck commentStamp: 'tpr 12/16/2016 18:06' prior: 0!
A MQTTPacketUnsubAck is returned by the server after an UNSUBSCRIBE. See section 3.11

The variable header is just the msgID provided by the initiating UNSUBSCRIBE packet (3.11.2). There is no payload. An UNSUBSCRIBE that lists several topics is responded to with a single UNSUBACK (3.10.4). We get the UNSUBACK even if no topics are actually cancelled.!

!MQTTPacketUnsubAck methodsFor: 'handling' stamp: 'tpr 1/27/2017 16:07'!
evaluateFor: anMQTTClient
	"The broker has responded to an UNSUB from me"
	anMQTTClient handleUnsubAckPacket: self! !

!MQTTPacketUnsubAck methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:38'!
packetType
	^11! !

!MQTTPacketVariableHeaded methodsFor: 'encoding' stamp: 'tpr 12/24/2016 17:35'!
decodeFrom: aReadStream
	"decode my fixed header and then my variable header"
		super decodeFrom: aReadStream.
		self decodeVariableHeaderFrom: aReadStream! !

!MQTTPacketVariableHeaded methodsFor: 'encoding' stamp: 'tpr 12/15/2016 16:21'!
decodeLengthFrom: aReadStream
"pull bytes from the stream and convert to the remainingLength value for this packet"
	| byte shift val |
	shift := 0.
	val := 0.
	[byte := aReadStream next.
		(byte isNil or: [shift > 21]) ifTrue:[self encodedLengthError].
		val := val + ((byte bitAnd: 127)<< shift).
		shift := shift + 7.
		(byte bitAnd: 128) > 0
	] whileTrue.
	remainingLength := val! !

!MQTTPacketVariableHeaded methodsFor: 'variable header' stamp: 'tpr 12/24/2016 17:15'!
decodeStringFrom: aReadStream
"read a string from the stream, remembering that it will be prepended with a 16 bit integer indicating the length of the string. Conver t from UTF8 to squeak character mapping."
	| ln string |
	ln := self decode16BitIntegerFrom: aReadStream.
	string := aReadStream next: ln.
	^string asString utf8ToSqueak! !

!MQTTPacketVariableHeaded methodsFor: 'variable header' stamp: 'tpr 12/22/2016 16:44'!
decodeVariableHeaderFrom: aReadStream
"read the variable header from the stream. Not all subclasses actually use msgIDs, so be careful to catch those"
	msgID := self decode16BitIntegerFrom: aReadStream! !

!MQTTPacketVariableHeaded methodsFor: 'encoding' stamp: 'tpr 12/19/2016 21:07'!
encodeFixedHeaderOn: aWriteStream
"write the fixed header on the stream"
	aWriteStream nextPut: self fixedHeader.
	self encodeLengthOn: aWriteStream! !

!MQTTPacketVariableHeaded methodsFor: 'encoding' stamp: 'tpr 12/15/2016 15:25'!
encodeLengthOn: aWriteStream
"encode the remaining length in the odd 7-bit manner used; see mqtt doc table 2.4 etc. write each byte to the stream"
	|val byte  |
	val := remainingLength.
	[byte := val \\ 128.
		val := val // 128.
		val > 0 ifTrue: [ byte := byte bitOr: 128].
		aWriteStream nextPut: byte.
		val >0
	] whileTrue! !

!MQTTPacketVariableHeaded methodsFor: 'encoding' stamp: 'tpr 12/26/2016 12:19'!
encodeOn: aWriteStream
	"when writing a packet out  we can either 
		write the fixed header
		calculate the remaining length (which involves doing almost all the work below) and writing that out
		write the variable header
		write the payload(s)
	or
		make a temporary stream
		write the variable header to the temp stream
		write the payload(s) tothe temp stream
		calculate the remaining length value from the size of the temp stream
		write the fixed header to the original stream
		write the calculated size to the original stream
		write the temp stream to the original stream
	I'm going to try the second method here. It might be even cleaner to use nextPutAllFlush: tempStream contents, but maybe later" 
	|tempStream|
	tempStream := WriteStream on: (ByteArray new: self guessEncodedLength).
	self encodeVariableHeaderOn: tempStream.
	self encodePayloadOn: tempStream.
	remainingLength := tempStream position.
	self encodeFixedHeaderOn: aWriteStream.
	tempStream readStream  do: [:v| aWriteStream nextPut: v].
	aWriteStream flush.
	^aWriteStream! !

!MQTTPacketVariableHeaded methodsFor: 'encoding' stamp: 'tpr 12/28/2016 13:58'!
encodePayloadOn: ignored
	"do nothing; subclasses that actually have payloads must do Their Own Thing"! !

!MQTTPacketVariableHeaded methodsFor: 'variable header' stamp: 'tpr 12/22/2016 16:41'!
encodeVariableHeaderOn: aWriteStream
"write the basic variable header to the stream. Not all subclasses actually use msgIDs, so be careful to catch those, and some subclass this method to add more data"
	self encode16BitInteger: msgID on: aWriteStream! !

!MQTTPacketVariableHeaded methodsFor: 'encoding' stamp: 'tpr 12/23/2016 16:31'!
guessEncodedLength
	"make a plausible guess at the final encoded length, erring on the side of excess"
	^100! !

!MQTTPacketVariableHeaded methodsFor: 'accessing' stamp: 'tpr 12/25/2016 12:50'!
packetID
	"return the packet identifier that many of my subclasses (but not all!!) contain"
	^msgID! !

!MQTTPacketVariableHeaded methodsFor: 'accessing' stamp: 'tpr 12/28/2016 11:51'!
packetID: a16BitNumber
	"set the packet identifier that many of my subclasses (but not all!!) contain.
	We ought to test this for validitiy"
	msgID := a16BitNumber! !

!MQTTPacketVariableHeaded methodsFor: 'printing' stamp: 'tpr 2/3/2017 13:00'!
printOn: aStream
	super printOn: aStream.
	msgID ifNotNil:[ :id| aStream nextPutAll: ' msgID: '.
		msgID printOn: aStream.
		aStream cr]
	! !

MQTTPacketVariableHeaded subclass: #MQTTPacketVariableHeadedWithPayload
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketVariableHeadedWithPayload commentStamp: 'tpr 12/16/2016 18:26' prior: 0!
MQTTPacketVariableHeadedWithPayload is an abstract class for the packets with both variable headers and payloads.
	CONNECT
	PUBLISH
	SUBSCRIBE
	SUBACK
	UNSUBSCRIBE!

MQTTPacketVariableHeadedWithPayload subclass: #MQTTPacketConnect
	instanceVariableNames: 'user password keepAliveTime willMessage willTopic willQos willRetainFlag cleanSessionFlag clientID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketConnect commentStamp: 'tpr 1/18/2017 16:10' prior: 0!
An MQTTPacketConnect contains all the data to make a CONNECT request per section 3.1 of the spec. 
We only send these; the server responds with a CONNACK. 

The variable header contains 4 fields to describe
a) protocol name (3.1.2.1) - a fixed 6 bytes of \0\4MQTT
b) protocol level (3.1.2.2) - currently a fixed value of \4 to signify MQTT 3.1.1 with an older version 3.1 using \3. If the returned CONNACK packer has a code of \1 it means the server does not understand the requested protocol version
c)connect flags (3.1.2.3) - a single byte encoding
userFlag | passwordFlag | willRetainedFlag | willQos (2 bits) | willFlag | cleanSessionFlag | reserverved to \0
d) keep alive time (3.1.2.10) - 2 bytes msb-first time in secondsipr\

Following the headers there is a payload containing
a) a client ID (3.1.3.1)
b) an optional Will topic (3.1.3.2) and WIll message (3.1.3.3)
c) an optional user name (3.1.3.4)
d) an optional password (3.1.3.5)

We expect to get a MQTTPacketConnAck back; only one will be sent and we need no matching id to test. Sending a second connect request is supposed to make the server close the connection. (3.1)
!

!MQTTPacketConnect methodsFor: 'initialize-release' stamp: 'tpr 1/26/2017 13:02'!
cleanSession: sessionBoolean
	"set the clean session flag; default is true"
	cleanSessionFlag := sessionBoolean! !

!MQTTPacketConnect methodsFor: 'initialize-release' stamp: 'tpr 1/18/2017 16:40'!
clientID: clientIDString
	"set the id that the server will use to identify the session being established by this connect packet. It ought to be unique within the scope of this image, ideally totally unique. That's up to the user of this class though"
	clientID := clientIDString! !

!MQTTPacketConnect methodsFor: 'payload' stamp: 'tpr 1/18/2017 16:41'!
clientIDString
	"ought to come from the client but if it hasn't been set use a vaguely useful default"
	^clientID ifNil:['SqueakMQTT']! !

!MQTTPacketConnect methodsFor: 'variable header' stamp: 'tpr 12/22/2016 16:57'!
connectFlags
	| byte |
	byte := willQos << 3.
	byte := byte bitAt: 2 put: cleanSessionFlag asBit.
	byte := byte bitAt: 3 put: willTopic notNil asBit.
	byte := byte bitAt: 6 put: willRetainFlag  asBit.
	byte := byte bitAt: 7 put: password notNil asBit.
	byte := byte bitAt: 8 put: user notNil asBit.
	^ byte! !

!MQTTPacketConnect methodsFor: 'variable header' stamp: 'tpr 12/24/2016 17:37'!
decodePayloadFrom: aReadStream
	"first read the CONNECT variable header (see 3.1.2) from the stream and then the actualy payload - purely for comparison and test purposes since we only send CONNECT packets. We do this odd order because the variable header includes flags that directly affect how the payload is interpreted"
	"first the 16 bit length of the protocol name"
	| cFlags willFlag passwordFlag userFlag|
	self assert: [(self decodeStringFrom: aReadStream) = 'MQTT'].
	self assert: [aReadStream next = MQTTProtocolLevel311].
	cFlags := aReadStream next.
	cleanSessionFlag := cFlags allMask: 2.
	willFlag := cFlags allMask: 4.
	willRetainFlag := cFlags allMask: 1<<5.
	willQos := cFlags >>3 bitAnd: 3.
	passwordFlag := cFlags allMask: 1<<6.
	userFlag := cFlags allMask: 1<<7.
	keepAliveTime := self decode16BitIntegerFrom: aReadStream.
	self assert: [(self decodeStringFrom: aReadStream) = self clientIDString].
	willFlag ifTrue: [
		willTopic := self decodeStringFrom: aReadStream.
		willMessage := self decodeStringFrom: aReadStream].
	userFlag ifTrue: [
		user := self decodeStringFrom: aReadStream].
	passwordFlag ifTrue: [
		password := self decodeStringFrom: aReadStream].
	self assert: [aReadStream position- 2 = remainingLength]
	
	
	! !

!MQTTPacketConnect methodsFor: 'variable header' stamp: 'tpr 12/24/2016 17:21'!
decodeVariableHeaderFrom: aReadStream
	"do nothing because the actual work is deferred to the decodePayloadFrom: method"! !

!MQTTPacketConnect methodsFor: 'payload' stamp: 'tpr 12/23/2016 17:12'!
encodePasswordOn: aWriteStream
	" if we have a username set, write it to the stream"
	password ifNotNil: [
		self encodeString: password on: aWriteStream]! !

!MQTTPacketConnect methodsFor: 'payload' stamp: 'tpr 12/23/2016 10:47'!
encodePayloadOn: aWriteStream
	"write my payload onto the stream- 
	the client ID
	iff it exists, the will topic
		the will message
	iff it exists, the username
	iff it exists, the password"
	self encodeString: self clientIDString on: aWriteStream.
	self encodeWillDataOn: aWriteStream.
	self encodeUsernameOn: aWriteStream.
	self encodePasswordOn: aWriteStream
	! !

!MQTTPacketConnect methodsFor: 'payload' stamp: 'tpr 12/23/2016 10:57'!
encodeUsernameOn: aWriteStream
	" if we have a username set, write it to the stream"
	user ifNotNil: [
		self encodeString: user on: aWriteStream]! !

!MQTTPacketConnect methodsFor: 'variable header' stamp: 'tpr 12/23/2016 17:02'!
encodeVariableHeaderOn: aWriteStream
"write the CONNECT variable header (see 3.1.2) to the stream"
	self encodeString: 'MQTT' on: aWriteStream.
	aWriteStream nextPut: MQTTProtocolLevel311 ; nextPut: self connectFlags.
	self encode16BitInteger: keepAliveTime on: aWriteStream
	
	! !

!MQTTPacketConnect methodsFor: 'payload' stamp: 'tpr 12/23/2016 10:56'!
encodeWillDataOn: aWriteStream
	"If I have a will set up, write the topic and then the message to the stream"
	willTopic ifNotNil:[
		self encodeString: willTopic on: aWriteStream.
		willMessage 
			ifNil: [self encodeString: '' on: aWriteStream]
			ifNotNil:[ self encodeString: willMessage on: aWriteStream]
		]
	! !

!MQTTPacketConnect methodsFor: 'initialize-release' stamp: 'tpr 1/18/2017 16:47'!
initialize
	"set the basic state to no user/password/will stuff/clean, just for safety. Nil is fine for most of them but a couple of flags and values need more specific initial values."
	willRetainFlag := false.
	cleanSessionFlag := true. "normally we are establishing a new connection and don't want old data back"
	willQos := 0.
	keepAliveTime := 60
	! !

!MQTTPacketConnect methodsFor: 'initialize-release' stamp: 'tpr 12/23/2016 15:51'!
keepAliveTime: timeInSecs
	"set the keep alive time interval; default is 60"
	keepAliveTime := timeInSecs! !

!MQTTPacketConnect methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:36'!
packetType
	^1! !

!MQTTPacketConnect methodsFor: 'initialize-release' stamp: 'tpr 12/23/2016 15:49'!
user: userName password: passwd
	"setup the user name and password; it is possible to have either nil, apparently"
	user := userName.
	password := passwd! !

!MQTTPacketConnect methodsFor: 'will' stamp: 'tpr 12/23/2016 15:47'!
willTopic: topicString message: messageString retain: retainBoolean qos: qosValue
	"setup a will message. We must have a topic, the actual message, a retain flag and requested QOS"
	willTopic := topicString.
	willMessage := messageString.
	willRetainFlag := retainBoolean.
	willQos := qosValue min: 2 max: 0! !

MQTTPacketVariableHeadedWithPayload subclass: #MQTTPacketPublish
	instanceVariableNames: 'topic message duplicate qos retain'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketPublish commentStamp: 'tpr 12/16/2016 16:31' prior: 0!
An MQTTPacketPublish is used in both directions; we send one to publish information to a server and receive them to get information that we have subscribed to. See section 3.3 of the spec.

The fixed header contains 4 bits of data affecting how the packet is interpreted. These are encodings of the duplicate flag, the qos value and the duplicate flag; all instance variables. See 3.3.1 and #fixedHeaderFlags.
duplicate (3.3.1.1) - this indicates whether this is the first attempt to send this packet. True/1 indicates this may be a re-send.
qos (3.3.1.2 & table 3.2)- a value 0/1/2 (3 is reserved and must not be used) indicating the level of service required.
retain (3.3.1.3) - this indicates whether the message being published should be retained by the server and sent out to any new subscriber as a sort of initial state message

The variable header includes -
a) topic name (3.3.2.1) - (see section 1.5.3) a string of stuff
b) packet ID iff qos > 0

The payload is simply the data being published in an 'application specific format' which presumably includes text or binary, plain or encoded etc. A 0 length payload is acceptable, which may be a way to simply flag status in a minimal manner.

The recipient of a PUBLISH must respond according to the qos required from that packet (3.3.4)!

!MQTTPacketPublish methodsFor: 'pending jobs' stamp: 'tpr 2/3/2017 12:28'!
acknowledgement
	"Depending upon the qos we need to return a PUBACK or a PUBREC"
	qos = 0 ifTrue:[ ^self badQosError].
	qos = 1 ifTrue: [^MQTTPacketPubAck new packetID: msgID].
	^MQTTPacketPubRec new packetID: msgID ! !

!MQTTPacketPublish methodsFor: 'fixed header' stamp: 'tpr 12/15/2016 17:07'!
decodeFixedHeaderFrom: aReadStream
"read the fixed header from the stream and check it for decent state and extract the special flags"
	|hdrByte|
	(hdrByte := aReadStream next) >>4  = self packetType ifFalse:[ self badFixedHeaderError].
	duplicate := hdrByte allMask: 8.
	retain := hdrByte allMask: 1.
	qos := (hdrByte >> 1) bitAnd: 3.
	self decodeLengthFrom: aReadStream! !

!MQTTPacketPublish methodsFor: 'encoding' stamp: 'tpr 12/28/2016 17:45'!
decodeFrom: aReadStream
	"We need to work out the length of the variable header, which may have an arbitrary length utf8 string and packetID, in order to be able to work out how long the payload is"
	| ln |
	self decodeFixedHeaderFrom: aReadStream.
	ln := self decodeVariableHeaderFrom: aReadStream.
	message := aReadStream next: (remainingLength - ln)
	! !

!MQTTPacketPublish methodsFor: 'variable header' stamp: 'tpr 12/28/2016 21:12'!
decodeTopicStringFrom: aReadStream
"read a string from the stream, remembering that it will be prepended with a 16 bit integer indicating the length of the string. Convert from UTF8 to squeak character mapping.
Set the string as my topic and return the (encoded) string length to help in decoding the payload.
This replace a simple #decodeStringFrom: normally used"
	| ln string |
	ln := self decode16BitIntegerFrom: aReadStream.
	string := aReadStream next: ln.
	topic := string asString utf8ToSqueak.
	^ln +2! !

!MQTTPacketPublish methodsFor: 'variable header' stamp: 'tpr 12/28/2016 17:44'!
decodeVariableHeaderFrom: aReadStream
	"extract the topic string and then iff qos >0, the msgID. Return the (encoded) length of the variable header to help in decoding the payload"
	| ln |
	ln := self decodeTopicStringFrom: aReadStream.
	qos > 0 ifTrue: [super decodeVariableHeaderFrom: aReadStream. ln := ln +2].
	^ln! !

!MQTTPacketPublish methodsFor: 'accessing' stamp: 'tpr 12/26/2016 17:53'!
duplicateFlag: aBoolean
	"set my DUP flag; this needs to be done by the client after the first send attempt for a qos>0 packet"
	duplicate := aBoolean! !

!MQTTPacketPublish methodsFor: 'variable header' stamp: 'tpr 12/26/2016 17:48'!
encodePayloadOn: aWriteStream
	"PUBLISH packets put the message data out as is with no prepended size field. We work out the data size from the remainingLength value etc"
	 aWriteStream nextPutAll: message! !

!MQTTPacketPublish methodsFor: 'variable header' stamp: 'tpr 12/26/2016 17:45'!
encodeVariableHeaderOn: aWriteStream
	"PUBLISH packets put the topic into the variable header followed by a msgID iff qos >0"
	topic ifNil: [^self badTopicError].
	self encodeString: topic on: aWriteStream.
	qos > 0 ifTrue:[ super encodeVariableHeaderOn: aWriteStream]! !

!MQTTPacketPublish methodsFor: 'handling' stamp: 'tpr 2/3/2017 12:39'!
evaluateFor: anMQTTClient
	"The broker has sent a PUBLISH packet to me"
	anMQTTClient handlePublishPacket: self qos: qos! !

!MQTTPacketPublish methodsFor: 'fixed header' stamp: 'tpr 12/26/2016 22:16'!
fixedHeaderFlags
"a publish packet is the only one with flags dependent upon variable state
DUP - duplicate delivery
QoS - Quality of Service, 2 bits
RETAIN - Retained message flag"

	^((duplicate ifTrue:[8] ifFalse:[0]) bitOr: (qos << 1)) bitOr: (retain ifTrue:[1] ifFalse:[0])

! !

!MQTTPacketPublish methodsFor: 'initialize-release' stamp: 'tpr 12/26/2016 17:35'!
initialize
	"set default flags & qos"
	qos := 0.
	duplicate := retain := false! !

!MQTTPacketPublish methodsFor: 'handling' stamp: 'tpr 2/7/2017 17:45'!
matchesSubscription: subscription ifTrue: aBlock
	"does my topic match the pattern? If so, evaluate the block"
	^(subscription topicMatches: topic) ifTrue: [ aBlock value: topic value: message]! !

!MQTTPacketPublish methodsFor: 'payload' stamp: 'tpr 12/28/2016 17:48'!
messageString
	"return the message data asa String"
	^ message asString utf8ToSqueak! !

!MQTTPacketPublish methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:37'!
packetType
	^3! !

!MQTTPacketPublish methodsFor: 'pending jobs' stamp: 'tpr 2/4/2017 17:09'!
pendingJob
	"Depending upon the qos we need to shedule a pending job for a PUBACK or a PUBREC"
	qos = 0 ifTrue:[ ^self badQosError].
	qos = 1 ifTrue: [^MQTTPendingPubAckJob for: self].
	^MQTTPendingPubRecJob for: self! !

!MQTTPacketPublish methodsFor: 'pending jobs' stamp: 'tpr 2/4/2017 17:09'!
pendingReleaseJob
	"We are handling a PUBLISH with qos=2 and need to schedule apending handler"
	^MQTTPendingPubRelJob for: self! !

!MQTTPacketPublish methodsFor: 'accessing' stamp: 'tpr 2/6/2017 13:38'!
prepareForResend
	"set my DUP flag; this needs to be done by the client after the first send attempt for a qos>0 packet"
	duplicate := true! !

!MQTTPacketPublish methodsFor: 'printing' stamp: 'tpr 2/3/2017 12:19'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' topic '.
	topic asString printOn: aStream.
	aStream nextPutAll: ' message: '.
	message asString printOn: aStream.
	aStream nextPutAll: ' qos: '.
	qos asString printOn: aStream.
	aStream cr! !

!MQTTPacketPublish methodsFor: 'accessing' stamp: 'tpr 12/26/2016 17:51'!
qos: qosValue
	"set my qos requested value 0-2"
	qos := qosValue min: 2 max: 0! !

!MQTTPacketPublish methodsFor: 'accessing' stamp: 'tpr 12/26/2016 22:21'!
retainFlag: aBoolean
	"set my RETAIN flag; this is supposed to be set if you want the message to be held by the broker for broadcasting in response to new subscribes."
	retain := aBoolean! !

!MQTTPacketPublish methodsFor: 'accessing' stamp: 'tpr 7/31/2017 16:06'!
topic: topicString message: messageBytes
	"set my topic string and message data.
	The topic must be a plain Squeak String which we will now check for any invalid characters and convert to UTF8. The messageBytes must be a ByteArray"

	topic := topicString squeakToUtf8 asByteArray.
	(topic size > 65535 or: [topic includesAnyOf: #($# $/)]) ifTrue: [^self badTopicError].

	messageBytes class = ByteArray ifFalse: [^self badMessageError]. "I hate to use such specific test but it seems important here"
	message := messageBytes! !

MQTTPacketVariableHeadedWithPayload subclass: #MQTTPacketSubAck
	instanceVariableNames: 'returnCodes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketSubAck commentStamp: 'tpr 12/16/2016 17:56' prior: 0!
A MQTTPacketSubAck is the response from the server to a SUBSCRIBE packet. We only receive these. See section 3.9

The variable header consists of the same msgID as the initiating SUBSCRIBE.

The payload is a list of return codes in the same order as the originallly SUBSCRIBEd topics. Each code provides the maximum qos the server will allow for each topic, or \128 for a failure (3.9.3)!

!MQTTPacketSubAck methodsFor: 'payload' stamp: 'tpr 12/28/2016 14:39'!
decodePayloadFrom: aReadStream
	"read my payload from the stream"
	returnCodes := aReadStream next: (remainingLength -2)! !

!MQTTPacketSubAck methodsFor: 'payload' stamp: 'tpr 1/27/2017 16:12'!
evaluateFor: anMQTTClient
	"The broker has responded to a SUBSCRIBE from me"
	anMQTTClient handleSubAckPacket: self! !

!MQTTPacketSubAck methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:37'!
packetType
	^9! !

MQTTPacketVariableHeadedWithPayload subclass: #MQTTPacketSubscribe
	instanceVariableNames: 'payloadDict'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketSubscribe commentStamp: 'tpr 12/16/2016 16:38' prior: 0!
A MQTTPacketSubscribe contains a request to the server for future information on a topic or topics. We only send them. See section 3.8

The variable header consists of a msgID set by us (3.8.2). It will be used by the server iff qos > 0 and we are supposed to make sure msgIDs are not used by multiple active packets. This would seem to require some complex checking within lists of active packets, or other tag collections. The Python implementation doesn't seem to even pretend to do this right.

The payload is a list of topic filter strings (there must be at least one topic requested) accompanied by a requested qos (3.8.3).

The server must respond with a SUBACK having the same msgID.!

!MQTTPacketSubscribe methodsFor: 'accessing' stamp: 'tpr 12/28/2016 14:10'!
addTopic: aString qos: aSmallNumber
	"add this topic to the payload along with its qos request. Some checking for validity wouldn't be a bad idea"
	payloadDict at: aString put: (aSmallNumber min: 2 max: 0)! !

!MQTTPacketSubscribe methodsFor: 'encoding' stamp: 'tpr 12/28/2016 14:06'!
encodeOn: aWriteStream
	"check for having at least one topic and qos-request pair - fail if not"
	payloadDict ifEmpty: [^self badTopicListError].
	^super encodeOn: aWriteStream! !

!MQTTPacketSubscribe methodsFor: 'encoding' stamp: 'tpr 12/28/2016 14:14'!
encodePayloadOn: aWriteStream
	"write my payload onto the stream"
	payloadDict keysAndValuesDo: [:key :val|
		self encodeString: key on: aWriteStream.
		aWriteStream nextPut: (val bitAnd: 3)]! !

!MQTTPacketSubscribe methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:45'!
fixedHeaderFlags
	^2! !

!MQTTPacketSubscribe methodsFor: 'initialize-release' stamp: 'tpr 12/28/2016 14:07'!
initialize
	"I need a dictionary prepared for the payload(s)"
		payloadDict := Dictionary new:2 "keep initial guess small"! !

!MQTTPacketSubscribe methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:37'!
packetType
	^8! !

!MQTTPacketSubscribe methodsFor: 'pending jobs' stamp: 'tpr 1/23/2017 16:22'!
pendingJob
	"Return a pending SUBACK to complete later"
	^MQTTPendingSubAckJob new
		originalPacket: self ! !

MQTTPacketVariableHeadedWithPayload subclass: #MQTTPacketUnsubscribe
	instanceVariableNames: 'topics'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPacketUnsubscribe commentStamp: 'tpr 12/16/2016 18:03' prior: 0!
A MQTTPacketUnsubscribe is how we unsubscribe from topics and we only send them to the server. See section 3.10

The variable header is just a msgID that we provide. See also the MQTTPacketSubscribe class comment. (3.10.2)

The payload is a list of topics that we no longer wish to subscribe to (3.10.3); it does not have to exactly match subscribed topics and mismatched topics will be ignored. The server may complete delivery of data relating to now unsubscribed topics if there are already buffered or in-progress broadcasts such as qos>0 messages. (3.10.4).
!

!MQTTPacketUnsubscribe methodsFor: 'accessing' stamp: 'tpr 1/31/2017 12:53'!
addTopic: aTopic
	"add a topic to unsubscribe from"
	topics add: aTopic! !

!MQTTPacketUnsubscribe methodsFor: 'encoding' stamp: 'tpr 1/31/2017 12:53'!
encodeOn: aWriteStream
	"check for having at least one topic and qos-request pair - fail if not"
	topics ifEmpty: [^self badTopicListError].
	^super encodeOn: aWriteStream! !

!MQTTPacketUnsubscribe methodsFor: 'encoding' stamp: 'tpr 1/31/2017 12:53'!
encodePayloadOn: aWriteStream
	"write my payload onto the stream"
	topics do: [:val|
		self encodeString: val on: aWriteStream]! !

!MQTTPacketUnsubscribe methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:45'!
fixedHeaderFlags
	^2! !

!MQTTPacketUnsubscribe methodsFor: 'fixed header' stamp: 'tpr 1/31/2017 12:54'!
initialize

	topics := OrderedCollection new:2! !

!MQTTPacketUnsubscribe methodsFor: 'fixed header' stamp: 'tpr 12/13/2016 18:38'!
packetType
	^10! !

!MQTTPacketUnsubscribe methodsFor: 'pending jobs' stamp: 'tpr 2/4/2017 17:09'!
pendingJob
	"Return a pending UNSUBACK to complete later"
	^MQTTPendingUnsubAckJob for: self! !

!MQTTPacketVariableHeadedWithPayload methodsFor: 'encoding' stamp: 'tpr 12/20/2016 20:46'!
decodeFrom: aReadStream
	"decode my fixed header and then my variable header and then my payload"
		super decodeFrom: aReadStream.
		self decodePayloadFrom: aReadStream! !

!MQTTPacketVariableHeadedWithPayload methodsFor: 'payload' stamp: 'tpr 12/20/2016 20:46'!
decodePayloadFrom: aReadStream
	"read my payload from the stream"
	"default version does nothing yet"! !

!MQTTPacketVariableHeadedWithPayload methodsFor: 'payload' stamp: 'tpr 12/20/2016 20:45'!
encodePayloadOn: aWriteStream
	"write my payload onto the stream"
	"default version does nothing yet"! !

!MQTTPacketVariableHeadedWithPayload methodsFor: 'payload' stamp: 'tpr 12/23/2016 17:00'!
messageID: aNumber
	"set the message ID; may need to check validity of these"
	msgID := aNumber! !

Object subclass: #MQTTPendingJob
	instanceVariableNames: 'originalPacket sendTime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPendingJob commentStamp: 'tpr 2/6/2017 13:57' prior: 0!
MQTTPendingJobs are a way to record some response we are waiting to get from the broker.
For example, when we publish something at qos=1 we add a pending job for a PUBACK - an MQTTPendingPubAckJob.!

!MQTTPendingJob class methodsFor: 'instance creation' stamp: 'tpr 2/4/2017 17:00'!
for: anMQTTPacket
	"return a pending job for the packet"
	^self new originalPacket: anMQTTPacket! !

!MQTTPendingJob methodsFor: 'handling' stamp: 'tpr 2/24/2017 17:10'!
completeFor: anMQTTClient
	"we're done; close me down, release me and my ID to our rest"

	anMQTTClient releasePendingJob: self
! !

!MQTTPendingJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:50'!
isPubAckID: msgID
	"am I a pending PUBACK job for msgId?"
	^false! !

!MQTTPendingJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:50'!
isPubCompID: msgID
	"am I a pending PUBCOMP job for msgId?"
	^false! !

!MQTTPendingJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:51'!
isPubRecID: msgID
	"am I a pending PUBREC job for msgId?"
	^false! !

!MQTTPendingJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:51'!
isPubRelID: msgID
	"am I a pending PUBREL job for msgId?"
	^false! !

!MQTTPendingJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:51'!
isSubAckID: msgID
	"am I a pending SUBACK job for msgId?"
	^false! !

!MQTTPendingJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:51'!
isUnsubAckID: msgID
	"am I a pending UNSUBACK job for msgId?"
	^false! !

!MQTTPendingJob methodsFor: 'accessing' stamp: 'tpr 2/4/2017 17:22'!
originalPacket: mqttSubscribePacket
	"I need to remember the original subscribe data for later"
	originalPacket := mqttSubscribePacket.
	sendTime := Time totalSeconds "record the last time it was sent"! !

!MQTTPendingJob methodsFor: 'accessing' stamp: 'tpr 2/4/2017 16:58'!
packetID
	^originalPacket packetID! !

!MQTTPendingJob methodsFor: 'resending' stamp: 'tpr 2/4/2017 17:30'!
resendFor: anMQTTClient ifNeededAtTime: seconds
	"If seconds (which is the actual time - the retryTime currently in use) > my last sendTime,  resend my original packet with any required changes and update that sendTime"
	
	seconds >= sendTime ifTrue:[
		sendTime := Time totalSeconds.
		anMQTTClient sendPacket: originalPacket]! !

MQTTPendingJob subclass: #MQTTPendingPingJob
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPendingPingJob commentStamp: 'tpr 2/6/2017 15:09' prior: 0!
A MQTTPendingPingJob is a bit of a hack to support the MQTT keep-alive pings. By adding a pending job for a PINGREQ we can tap into the regular cycle of testing for packets needing a resend; the ping job pushes out a ping and resets it's next-required time. !

!MQTTPendingPingJob methodsFor: 'handling' stamp: 'tpr 2/6/2017 17:17'!
completeFor: anMQTTClient
	"I do nothing here; I'm so very special"
! !

!MQTTPendingPingJob methodsFor: 'initialize-release' stamp: 'tpr 2/6/2017 16:28'!
initialize
	sendTime := Time totalSeconds! !

!MQTTPendingPingJob methodsFor: 'accessing' stamp: 'tpr 3/2/2017 14:01'!
packetID
	"return a too-big value for real packets so the #newPacketID check never 'finds' this"
	^16r10000! !

!MQTTPendingPingJob methodsFor: 'resending' stamp: 'tpr 2/10/2017 18:23'!
resendFor: anMQTTClient ifNeededAtTime: seconds
	"If seconds (which is the actual time - the retryTime currently in use) > my last sendTime,  tell the client to ping and update the sendTime. #ping returns a good time to re-ping"

	seconds >= sendTime ifTrue:[
		sendTime := anMQTTClient ping]! !

MQTTPendingJob subclass: #MQTTPendingPubAckJob
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPendingPubAckJob commentStamp: 'tpr 2/6/2017 13:58' prior: 0!
A PUBLISH packet with qos=1 has been sent and we expect a corresponding PUBACK sometime soon.
An MQTTPendingPubAckJob is used to record that expectation; it also holds the original PUBLISH in case we need to re-send it - and we set the DUP flag to true when we get the original packet in order to be ready.!

!MQTTPendingPubAckJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:52'!
isPubAckID: msgID
	"am I a pending PUBACK job for msgId?"
	^self packetID = msgID! !

!MQTTPendingPubAckJob methodsFor: 'accessing' stamp: 'tpr 2/6/2017 13:46'!
originalPacket: mqttSubscribePacket
	"I need to set the DUP bit in case of a resend"
	super originalPacket: mqttSubscribePacket.
	mqttSubscribePacket prepareForResend! !

MQTTPendingJob subclass: #MQTTPendingPubCompJob
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPendingPubCompJob commentStamp: 'tpr 2/6/2017 14:07' prior: 0!
A PUBLISH packet with qos=2 has been sent, the PUBREC has arrived, we have sent the PUBREL and we expect a corresponding PUBCOMP soon.
An MQTTPendingPubCompJob is used to record that expectation; it also holds the original PUBREL in case we need to re-send it. Sending the PUBCOMP is the end of the dance.!

!MQTTPendingPubCompJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:53'!
isPubCompID: msgID
	"am I a pending PUBCOMP job for msgId?"
	^self packetID = msgID! !

MQTTPendingJob subclass: #MQTTPendingPubRecJob
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPendingPubRecJob commentStamp: 'tpr 2/6/2017 14:02' prior: 0!
A PUBLISH packet with qos=2 has been sent and we expect a corresponding PUBREC sometime soon.
An MQTTPendingPubRecJob is used to record that expectation; it also holds the original PUBLISH in case we need to re-send it - and we set the DUP flag to true when we get the original packet in order to be ready. Once we get the PUBREC we send back a PUBREL packet and do a little dance!

!MQTTPendingPubRecJob methodsFor: 'handling' stamp: 'tpr 2/6/2017 17:06'!
completeFor: anMQTTClient
	"we're done; close me down, release me and *not* my ID to our rest"

	anMQTTClient releasePendingJob: self
! !

!MQTTPendingPubRecJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:54'!
isPubRecID: msgID
	"am I a pending PUBREC job for msgId?"
	^self packetID = msgID! !

!MQTTPendingPubRecJob methodsFor: 'accessing' stamp: 'tpr 2/6/2017 13:46'!
originalPacket: mqttSubscribePacket
	"I need to set the DUP bit in case of a resend"
	super originalPacket: mqttSubscribePacket.
	mqttSubscribePacket prepareForResend! !

MQTTPendingJob subclass: #MQTTPendingPubRelJob
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPendingPubRelJob commentStamp: 'tpr 2/6/2017 14:05' prior: 0!
A PUBLISH packet with qos=2 has been receieved and we have sent the appropriate PUBREC acknowledgment and now we expect a corresponding PUBREL sometime.
An MQTTPendingPubRelJob is used to record that expectation; it also holds the original PUBREC in case we need to re-send it. Once we get the PUBREL we send back a PUBCOMP packet and conclude our little dance!

!MQTTPendingPubRelJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:55'!
isPubRelID: msgID
	"am I a pending PUBREL job for msgId?"
	^self packetID = msgID! !

MQTTPendingJob subclass: #MQTTPendingSubAckJob
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPendingSubAckJob commentStamp: 'tpr 2/6/2017 14:14' prior: 0!
A MQTTPendingSubAckJob is how we wait for an acknowledgement of a SUBSCRIBE. We need to check the returned error codes for each subscription we sent.

It looks like we don't resend these.!

!MQTTPendingSubAckJob methodsFor: 'handling' stamp: 'tpr 2/6/2017 17:13'!
completeFor: anMQTTClient
	"we're done; close me down, release me and my ID to our rest"
	super completeFor: anMQTTClient.
	
	"Now do the checking of the return codes"
	Transcript show: 'suback return codes need handling'; cr! !

!MQTTPendingSubAckJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:55'!
isSubAckID: msgID
	"am I a pending SUBACK job for msgId?"
	^self packetID = msgID! !

!MQTTPendingSubAckJob methodsFor: 'resending' stamp: 'tpr 2/6/2017 14:14'!
resendFor: anMQTTClient ifNeededAtTime: seconds
	"do nothing for this packet"! !

MQTTPendingJob subclass: #MQTTPendingUnsubAckJob
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTPendingUnsubAckJob commentStamp: 'tpr 2/6/2017 14:14' prior: 0!
A MQTTPendingUnsubAckJob is how we wait for an acknowledgement of an UNSUBSCRIBE. 

It looks like we don't resend these.!

!MQTTPendingUnsubAckJob methodsFor: 'testing' stamp: 'tpr 2/6/2017 16:57'!
isUnsubAckID: msgID
	"am I a pending UNSUBACK job for msgId?"
	^self packetID = msgID! !

!MQTTPendingUnsubAckJob methodsFor: 'resending' stamp: 'tpr 2/6/2017 14:14'!
resendFor: anMQTTClient ifNeededAtTime: seconds
	"do nothing for this packet"! !

Object subclass: #MQTTSubscription
	instanceVariableNames: 'topic actionBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTSubscription commentStamp: 'tpr 2/24/2017 17:25' prior: 0!
A MQTTSubscription is how we keep track of subscribed topics and the action the user wants when a packet comes in

topic - a String that we match against an incoming packet to see if we need to do anything. See the MQTT documentation for the rules about wildcards

actionBlock - the Block we run if we match the topic; it gets two parameters, the topic string and the message data!

!MQTTSubscription class methodsFor: 'as yet unclassified' stamp: 'tpr 2/7/2017 15:25'!
for: aTopic do: aBlock

	^super new for: aTopic do: aBlock! !

!MQTTSubscription methodsFor: 'initialize-release' stamp: 'tpr 2/24/2017 14:28'!
for: aTopic do: aBlock 
	"build a response to data for aTopic being received; check the validity of aTopic (no nulls, + or # used properly etc) and 
	return nil if there is a problem, otherwise return self"
	| strm priorChar currentChar nextChar count |
	topic := aTopic.
	actionBlock := aBlock.
	priorChar := nil.
	count := 0.
	strm := aTopic readStream.
	[currentChar := strm next]
		whileNotNil: [nextChar := strm peek.
			currentChar = $+
				ifTrue: [((priorChar notNil and: [priorChar ~= $/]) or: [nextChar notNil and: [nextChar ~= $/]])
						ifTrue: [^ nil]]
				ifFalse: [currentChar = $#
						ifTrue: [((priorChar notNil and: [priorChar ~= $/]) or: [nextChar notNil])
								ifTrue: [^ nil]]
						ifFalse: [currentChar = Character null
								ifTrue: [^ nil]]].
			count := count + 1.
			priorChar := currentChar].
	count > 65535
		ifTrue: [^ nil]! !

!MQTTSubscription methodsFor: 'handling packets' stamp: 'tpr 2/7/2017 17:42'!
handlePacket: aPublishPacket
	"see if the packet's topic matches mine; if so, evaluate the actionBlock"

	^aPublishPacket matchesSubscription: self ifTrue: actionBlock! !

!MQTTSubscription methodsFor: 'handling packets' stamp: 'tpr 2/13/2017 18:33'!
topicMatches: candidateString 
	"does my topic match this topic? First version simply does a string =
	test, needs to actually parse the pattern properly"
	"^topic sameAs: candidateString"
	"copied from Python and C. Ugly code. Fix!!
	See git.eclipse.org/mosquitto/org.eclipse.mosquitto.git/tree/llib/util_mosq.c for the original"
	| result multilevel topicLen csLen topicPos csPos |
	result := true.
	multilevel := false.
	topicLen := topic size.
	csLen := candidateString size.
	(topicLen > 0 and: [csLen > 0])
		ifTrue: [(topic first = $$ xor: candidateString first = $$)
				ifTrue: [^ false]]. "testing for both or neither first char being $$"
			
	topicPos := csPos := 1.
	
	[topicPos <= topicLen and: [csPos <= csLen]] whileTrue: [
			(topic at: topicPos) = (candidateString at: csPos)
				ifTrue: [csPos = csLen
						ifTrue: ["check for topic ending in /#"
							(topicPos = (topicLen - 2) and: [(topic last: 2) = '/#']) ifTrue: [^ true]].
					"move to next char"
					topicPos := topicPos + 1.
					csPos := csPos + 1.
					(csPos > csLen and: [topicPos > topicLen])
						ifTrue: [^ true]
						ifFalse: [(csPos > csLen and: [topicPos = topicLen] and: [topic last = $+])
								ifTrue: [^ true]]]
				ifFalse: [(topic at: topicPos) = $+
						ifTrue: [topicPos := topicPos + 1.
							[csPos <= csLen and: [(candidateString at: csPos) ~= $/]] whileTrue: [csPos := csPos + 1].
							(csPos > csLen and: [topicPos > topicLen])
								ifTrue: [^ true]]
						ifFalse: [(topic at: topicPos) = $#
								ifTrue: [multilevel := true.
									^ topicPos = topicLen]
								ifFalse: [^ false]]]
			].
		(multilevel not and:[csPos <= csLen or:[topicPos <= topicLen]]) ifTrue: [^false].
		^result! !

Error subclass: #MQTTBrokerLookupFailure
	instanceVariableNames: 'brokerName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTBrokerLookupFailure class methodsFor: 'instance creation' stamp: 'tpr 3/15/2017 17:04'!
brokerName: aString
	^super new brokerName: aString! !

!MQTTBrokerLookupFailure methodsFor: 'handling' stamp: 'tpr 3/15/2017 16:42'!
brokerName

	^brokerName! !

!MQTTBrokerLookupFailure methodsFor: 'handling' stamp: 'tpr 3/15/2017 16:42'!
brokerName: aString

	brokerName:= aString! !

Error subclass: #MQTTConnectionException
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MQTT-Client'!

!MQTTConnectionException methodsFor: 'handling' stamp: 'tpr 3/15/2017 13:41'!
defaultAction

	self resume! !

!MQTTConnectionException methodsFor: 'exceptionDescription' stamp: 'tpr 3/15/2017 13:41'!
isResumable

	^true! !
MQTTClient initialize!
MQTTPacket initialize!
